#!/usr/bin/python
"""Tiny HTTP Proxy.

This module implements GET, HEAD, POST, PUT and DELETE methods
on BaseHTTPServer, and behaves as an HTTP proxy.  The CONNECT
method is also implemented experimentally, but has not been
tested yet.

Any help will be greatly appreciated.       SUZUKI Hisao

2009/11/23 - Modified by Mitko Haralanov
             * Added very simple FTP file retrieval
             * Added custom logging methods
             * Added code to make this a standalone application
"""
import ftplib
import getopt
import logging.handlers
import os
import select
import signal
import socket
import sys
import threading
from time import sleep
from types import CodeType
from types import FrameType

from caldav.lib.python_utilities import to_local
from caldav.lib.python_utilities import to_wire
from six import PY3

if PY3:
    from urllib import parse
    from urllib.parse import urlparse, urlunparse
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from socketserver import ThreadingMixIn
else:
    from urlparse import urlparse as parse
    from urlparse import urlparse, urlunparse
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
    from SocketServer import ThreadingMixIn

__version__ = "0.3.1"

DEFAULT_LOG_FILENAME = "proxy.log"


class ProxyHandler(BaseHTTPRequestHandler):
    __base = BaseHTTPRequestHandler
    __base_handle = __base.handle

    server_version = "TinyHTTPProxy/" + __version__
    rbufsize = 0  # self.rfile Be unbuffered

    def handle(self):
        (ip, port) = self.client_address
        self.server.logger.log(logging.INFO, "Request from '%s'", ip)
        if hasattr(self, "allowed_clients") and ip not in self.allowed_clients:
            self.raw_requestline = self.rfile.readline()
            if self.parse_request():
                self.send_error(403)
        else:
            self.__base_handle()

    def _connect_to(self, netloc, soc):
        i = netloc.find(":")
        if i >= 0:
            host_port = netloc[:i], int(netloc[i + 1 :])
        else:
            host_port = netloc, 80
        self.server.logger.log(
            logging.INFO, "connect to %s:%d", host_port[0], host_port[1]
        )
        try:
            soc.connect(host_port)
        except socket.error as arg:
            try:
                msg = arg[1]
            except:
                msg = arg
            self.send_error(404, msg)
            return 0
        return 1

    def do_CONNECT(self):
        soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            if self._connect_to(self.path, soc):
                self.log_request(200)
                self.wfile.write(
                    self.protocol_version + " 200 Connection established\r\n"
                )
                self.wfile.write("Proxy-agent: %s\r\n" % self.version_string())
                self.wfile.write("\r\n")
                self._read_write(soc, 300)
        finally:
            soc.close()
            self.connection.close()

    def do_GET(self):
        (scm, netloc, path, params, query, fragment) = urlparse(self.path, "http")
        if scm not in ("http", "ftp") or fragment or not netloc:
            self.send_error(400, "bad url %s" % self.path)
            return
        soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            if scm == "http":
                if self._connect_to(netloc, soc):
                    self.log_request()
                    soc.send(
                        to_wire(
                            "%s %s %s\r\n"
                            % (
                                self.command,
                                urlunparse(("", "", path, params, query, "")),
                                self.request_version,
                            )
                        )
                    )
                    self.headers["Connection"] = "close"
                    del self.headers["Proxy-Connection"]
                    for key_val in list(self.headers.items()):
                        soc.send(to_wire("%s: %s\r\n" % key_val))
                    soc.send(to_wire("\r\n"))
                    self._read_write(soc)
            elif scm == "ftp":
                # fish out user and password information
                i = netloc.find("@")
                if i >= 0:
                    login_info, netloc = netloc[:i], netloc[i + 1 :]
                    try:
                        user, passwd = login_info.split(":", 1)
                    except ValueError:
                        user, passwd = "anonymous", None
                else:
                    user, passwd = "anonymous", None
                self.log_request()
                try:
                    ftp = ftplib.FTP(netloc)
                    ftp.login(user, passwd)
                    if self.command == "GET":
                        ftp.retrbinary("RETR %s" % path, self.connection.send)
                    ftp.quit()
                except Exception as e:
                    self.server.logger.log(logging.WARNING, "FTP Exception: %s", e)
        finally:
            soc.close()
            self.connection.close()

    def _read_write(self, soc, max_idling=20, local=False):
        iw = [self.connection, soc]
        local_data = ""
        ow = []
        count = 0
        while 1:
            count += 1
            (ins, _, exs) = select.select(iw, ow, iw, 1)
            if exs:
                break
            if ins:
                for i in ins:
                    if i is soc:
                        out = self.connection
                    else:
                        out = soc
                    data = i.recv(8192)
                    if data:
                        if local:
                            local_data += data
                        else:
                            out.send(data)
                        count = 0
            if count == max_idling:
                break
        if local:
            return to_local(local_data)
        return None

    do_HEAD = do_GET
    do_POST = do_GET
    do_PUT = do_GET
    do_DELETE = do_GET
    do_PROPFIND = do_GET

    def log_message(self, format, *args):
        self.server.logger.log(
            logging.INFO, "%s %s", self.address_string(), format % args
        )

    def log_error(self, format, *args):
        self.server.logger.log(
            logging.ERROR, "%s %s", self.address_string(), format % args
        )


class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
    def __init__(self, server_address, RequestHandlerClass, logger=None):
        HTTPServer.__init__(self, server_address, RequestHandlerClass)
        self.logger = logger


class NonThreadingHTTPServer(HTTPServer):
    def __init__(self, server_address, RequestHandlerClass, logger=None):
        HTTPServer.__init__(self, server_address, RequestHandlerClass)
        self.logger = logger


def logSetup(filename, log_size, daemon):
    logger = logging.getLogger("TinyHTTPProxy")
    logger.setLevel(logging.INFO)
    if not filename:
        if not daemon:
            # display to the screen
            handler = logging.StreamHandler()
        else:
            handler = logging.handlers.RotatingFileHandler(
                DEFAULT_LOG_FILENAME, maxBytes=(log_size * (1 << 20)), backupCount=5
            )
    else:
        handler = logging.handlers.RotatingFileHandler(
            filename, maxBytes=(log_size * (1 << 20)), backupCount=5
        )
    fmt = logging.Formatter(
        "[%(asctime)-12s.%(msecs)03d] "
        "%(levelname)-8s {%(name)s %(threadName)s}"
        " %(message)s",
        "%Y-%m-%d %H:%M:%S",
    )
    handler.setFormatter(fmt)

    logger.addHandler(handler)
    return logger


def usage(msg=None):
    if msg:
        print(msg)
    print(sys.argv[0], "[-p port] [-l logfile] [-dh] [allowed_client_name ...]]")
    print()
    print("   -p       - Port to bind to")
    print("   -l       - Path to logfile. If not specified, STDOUT is used")
    print("   -d       - Run in the background")
    print()


def handler(signo, frame):
    while frame and isinstance(frame, FrameType):
        if frame.f_code and isinstance(frame.f_code, CodeType):
            if "run_event" in frame.f_code.co_varnames:
                frame.f_locals["run_event"].set()
                return
        frame = frame.f_back


def daemonize(logger):
    class DevNull(object):
        def __init__(self):
            self.fd = os.open("/dev/null", os.O_WRONLY)

        def write(self, *args, **kwargs):
            return 0

        def read(self, *args, **kwargs):
            return 0

        def fileno(self):
            return self.fd

        def close(self):
            os.close(self.fd)

    class ErrorLog:
        def __init__(self, obj):
            self.obj = obj

        def write(self, string):
            self.obj.log(logging.ERROR, string)

        def read(self, *args, **kwargs):
            return 0

        def close(self):
            pass

    if os.fork() != 0:
        # allow the child pid to instanciate the server
        # class
        sleep(1)
        sys.exit(0)
    os.setsid()
    fd = os.open("/dev/null", os.O_RDONLY)
    if fd != 0:
        os.dup2(fd, 0)
        os.close(fd)
    null = DevNull()
    log = ErrorLog(logger)
    sys.stdout = null
    sys.stderr = log
    sys.stdin = null
    fd = os.open("/dev/null", os.O_WRONLY)
    # if fd != 1: os.dup2(fd, 1)
    os.dup2(sys.stdout.fileno(), 1)
    if fd != 2:
        os.dup2(fd, 2)
    if fd not in (1, 2):
        os.close(fd)


def main():
    logfile = None
    daemon = False
    max_log_size = 20
    port = 8080
    allowed = []
    run_event = threading.Event()
    local_hostname = socket.gethostname()

    try:
        opts, args = getopt.getopt(sys.argv[1:], "l:dhp:", [])
    except getopt.GetoptError as e:
        usage(str(e))
        return 1

    for opt, value in opts:
        if opt == "-p":
            port = int(value)
        if opt == "-l":
            logfile = value
        if opt == "-d":
            daemon = not daemon
        if opt == "-h":
            usage()
            return 0

    # setup the log file
    logger = logSetup(logfile, max_log_size, daemon)

    if daemon:
        daemonize(logger)
    signal.signal(signal.SIGINT, handler)

    if args:
        allowed = []
        for name in args:
            client = socket.gethostbyname(name)
            allowed.append(client)
            logger.log(logging.INFO, "Accept: %s (%s)" % (client, name))
        ProxyHandler.allowed_clients = allowed
    else:
        logger.log(logging.INFO, "Any clients will be served...")

    server_address = (socket.gethostbyname(local_hostname), port)
    ProxyHandler.protocol = "HTTP/1.0"
    httpd = ThreadingHTTPServer(server_address, ProxyHandler, logger)
    sa = httpd.socket.getsockname()
    print("Servering HTTP on", sa[0], "port", sa[1])
    req_count = 0
    while not run_event.isSet():
        try:
            httpd.handle_request()
            req_count += 1
            if req_count == 1000:
                logger.log(
                    logging.INFO,
                    "Number of active threads: %s",
                    threading.activeCount(),
                )
                req_count = 0
        except select.error as e:
            if e[0] == 4 and run_event.isSet():
                pass
            else:
                logger.log(logging.CRITICAL, "Errno: %d - %s", e[0], e[1])
    logger.log(logging.INFO, "Server shutdown")
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from caldav.lib.python_utilities import to_normal_str
from caldav.lib.python_utilities import to_unicode
from six import PY3

if PY3:
    from urllib.parse import (
        ParseResult,
        SplitResult,
        urlparse,
        unquote,
        quote,
        urlunparse,
    )
else:
    from urlparse import ParseResult, SplitResult
    from urlparse import urlparse, urlunparse
    from urllib import unquote, quote


def uc2utf8(input):
    # argh!  this feels wrong, but seems to be needed.
    if not PY3 and type(input) == unicode:
        return input.encode("utf-8")
    else:
        return input


class URL:
    """
    This class is for wrapping URLs into objects.  It's used
    internally in the library, end users should not need to know
    anything about this class.  All methods that accept URLs can be
    fed either with an URL object, a string or an urlparse.ParsedURL
    object.

    Addresses may be one out of three:

    1) a path relative to the DAV-root, i.e. "someuser/calendar" may
    refer to
    "http://my.davical-server.example.com/caldav.php/someuser/calendar".

    2) an absolute path, i.e. "/caldav.php/someuser/calendar"

    3) a fully qualified URL, i.e.
    "http://someuser:somepass@my.davical-server.example.com/caldav.php/someuser/calendar".
    Remark that hostname, port, user, pass is typically given when
    instantiating the DAVClient object and cannot be overridden later.

    As of 2013-11, some methods in the caldav library expected strings
    and some expected urlParseResult objects, some expected
    fully qualified URLs and most expected absolute paths.  The purpose
    of this class is to ensure consistency and at the same time
    maintaining backward compatibility.  Basically, all methods should
    accept any kind of URL.

    """

    def __init__(self, url):
        if isinstance(url, ParseResult) or isinstance(url, SplitResult):
            self.url_parsed = url
            self.url_raw = None
        else:
            self.url_raw = url
            self.url_parsed = None

    def __bool__(self):
        if self.url_raw or self.url_parsed:
            return True
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __eq__(self, other):
        if str(self) == str(other):
            return True
        # The URLs could have insignificant differences
        me = self.canonical()
        if hasattr(other, "canonical"):
            other = other.canonical()
        return str(me) == str(other)

    def __hash__(self):
        return hash(str(self))

    # TODO: better naming?  Will return url if url is already an URL
    # object, else will instantiate a new URL object
    @classmethod
    def objectify(self, url):
        if url is None:
            return None
        if isinstance(url, URL):
            return url
        else:
            return URL(url)

    # To deal with all kind of methods/properties in the ParseResult
    # class
    def __getattr__(self, attr):
        if "url_parsed" not in vars(self):
            raise AttributeError
        if self.url_parsed is None:
            self.url_parsed = urlparse(self.url_raw)
        if hasattr(self.url_parsed, attr):
            return getattr(self.url_parsed, attr)
        else:
            return getattr(self.__unicode__(), attr)

    # returns the url in text format
    def __str__(self):
        return to_normal_str(self.__unicode__())

    # returns the url in text format
    def __unicode__(self):
        if self.url_raw is None:
            self.url_raw = self.url_parsed.geturl()
        return to_unicode(self.url_raw)

    def __repr__(self):
        return "URL(%s)" % str(self)

    def strip_trailing_slash(self):
        if str(self)[-1] == "/":
            return URL.objectify(str(self)[:-1])
        else:
            return self

    def is_auth(self):
        return self.username is not None

    def unauth(self):
        if not self.is_auth():
            return self
        return URL.objectify(
            ParseResult(
                self.scheme,
                "%s:%s"
                % (self.hostname, self.port or {"https": 443, "http": 80}[self.scheme]),
                self.path.replace("//", "/"),
                self.params,
                self.query,
                self.fragment,
            )
        )

    def canonical(self):
        """
        a canonical URL ... remove authentication details, make sure there
        are no double slashes, and to make sure the URL is always the same,
        run it through the urlparser, and make sure path is properly quoted
        """
        url = self.unauth()

        arr = list(self.url_parsed)
        ## quoting path and removing double slashes
        arr[2] = quote(unquote(url.path.replace("//", "/")))
        ## sensible defaults
        if not arr[0]:
            arr[0] = "https"
        if arr[1] and not ":" in arr[1]:
            if arr[0] == "https":
                portpart = ":443"
            elif arr[0] == "http":
                portpart = ":80"
            else:
                portpart = ""
            arr[1] += portpart

        # make sure to delete the string version
        url.url_raw = urlunparse(arr)
        url.url_parsed = None

        return url

    def join(self, path):
        """
        assumes this object is the base URL or base path.  If the path
        is relative, it should be appended to the base.  If the path
        is absolute, it should be added to the connection details of
        self.  If the path already contains connection details and the
        connection details differ from self, raise an error.
        """
        pathAsString = str(path)
        if not path or not pathAsString:
            return self
        path = URL.objectify(path)
        if (
            (path.scheme and self.scheme and path.scheme != self.scheme)
            or (path.hostname and self.hostname and path.hostname != self.hostname)
            or (path.port and self.port and path.port != self.port)
        ):
            raise ValueError("%s can't be joined with %s" % (self, path))

        if path.path[0] == "/":
            ret_path = uc2utf8(path.path)
        else:
            sep = "/"
            if self.path.endswith("/"):
                sep = ""
            ret_path = "%s%s%s" % (self.path, sep, uc2utf8(path.path))
        return URL(
            ParseResult(
                self.scheme or path.scheme,
                self.netloc or path.netloc,
                ret_path,
                path.params,
                path.query,
                path.fragment,
            )
        )


def make(url):
    """Backward compatibility"""
    return URL.objectify(url)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
## YOU SHOULD MOST LIKELY NOT EDIT THIS FILE!
## Make a conf_private.py for personal configuration.
## Check conf_private.py.EXAMPLE
import logging

from caldav.davclient import DAVClient

# from .compability_issues import bedework, xandikos

####################################
# Import personal test server config
####################################

## TODO: there are probably more elegant ways of doing this?

try:
    from .conf_private import only_private  ## legacy compatibility

    test_public_test_servers = not only_private
except ImportError:
    try:
        from .conf_private import test_public_test_servers
    except ImportError:
        test_public_test_servers = False

try:
    from .conf_private import caldav_servers
except ImportError:
    caldav_servers = []
try:
    from .conf_private import test_private_test_servers

    if not test_private_test_servers:
        caldav_servers = []
except ImportError:
    pass

try:
    from .conf_private import xandikos_host, xandikos_port
except ImportError:
    xandikos_host = "localhost"
    xandikos_port = 8993  ## random port above 8000
try:
    from .conf_private import test_xandikos
except ImportError:
    try:
        import xandikos

        test_xandikos = True
    except:
        test_xandikos = False

try:
    from .conf_private import radicale_host, radicale_port
except ImportError:
    radicale_host = "localhost"
    radicale_port = 5232  ## default radicale host

try:
    from .conf_private import test_radicale
except ImportError:
    try:
        import radicale

        test_radicale = True
    except:
        test_radicale = False

try:
    from .conf_private import rfc6638_users
except ImportError:
    rfc6638_users = []

proxy = "127.0.0.1:8080"
proxy_noport = "127.0.0.1"

#####################
# Public test servers
#####################
## As of 2019-09, all of those are down.  Will try to fix Real Soon ... possibly before 2029 even.
if False:
    # if test_public_test_servers:

    ## TODO: this one is set up on emphemeral storage on OpenShift and
    ## then configured manually through the webui installer, it will
    ## most likely only work for some few days until it's down again.
    ## It's needed to hard-code the configuration into
    ## https://github.com/python-caldav/baikal

    caldav_servers.append(
        {
            "url": "http://baikal-caldav-servers.cloudapps.bitbit.net/html/cal.php/",
            "username": "baikaluser",
            "password": "asdf",
        }
    )

    # bedework:
    # * todos and journals are not properly supported -
    #   ref https://github.com/Bedework/bedework/issues/5
    # * propfind fails to return resourcetype,
    #   ref https://github.com/Bedework/bedework/issues/110
    # * date search on recurrences of recurring events doesn't work
    #   (not reported yet - TODO)
    caldav_servers.append(
        {
            "url": "http://bedework-caldav-servers.cloudapps.bitbit.net/ucaldav/",
            "username": "vbede",
            "password": "bedework",
            "incompatibilities": compatibility_issues.bedework,
        }
    )

    caldav_servers.append(
        {
            "url": "http://xandikos-caldav-servers.cloudapps.bitbit.net/",
            "username": "user1",
            "password": "password1",
            "incompatibilities": compatibility_issues.xandikos,
        }
    )

    # radicale
    caldav_servers.append(
        {
            "url": "http://radicale-caldav-servers.cloudapps.bitbit.net/",
            "username": "testuser",
            "password": "123",
            "nofreebusy": True,
            "nodefaultcalendar": True,
            "noproxy": True,
        }
    )

###################################################################
# Convenience - get a DAVClient object from the caldav_servers list
###################################################################
CONNKEYS = set(
    ("url", "proxy", "username", "password", "ssl_verify_cert", "ssl_cert", "auth")
)


def client(idx=None, **kwargs):
    if idx is None and not kwargs:
        return client(0)
    elif idx is not None and not kwargs and caldav_servers:
        return client(**caldav_servers[idx])
    elif not kwargs:
        return None
    for bad_param in (
        "incompatibilities",
        "backwards_compatibility_url",
        "principal_url",
        "enable",
    ):
        if bad_param in kwargs:
            kwargs.pop(bad_param)
    for kw in kwargs:
        if not kw in CONNKEYS:
            logging.critical(
                "unknown keyword %s in connection parameters.  All compatibility flags should now be sent as a separate list, see conf_private.py.EXAMPLE.  Ignoring."
                % kw
            )
            kwargs.pop(kw)
    return DAVClient(**kwargs)



from tests import compatibility_issues

## PRIVATE CALDAV SERVER(S) TO RUN TESTS TOWARDS
## Make a list of your own servers/accounts that you'd like to run the
## test towards.  Running the test suite towards a personal account
## should generally be safe, it should not mess up with content there
## and it should clean up after itself, but don't sue me if anything
## goes wrong ...

## Define your primary caldav server here
caldav_servers = [
    { # 0
        'url': "https://xandikos.tobixen.no/",
        "password": "kalenderfestøl",
        "username": "tobias",
        'incompatibilities': compatibility_issues.xandikos,
        'enable': True
    },
    { # 1
        'url': 'https://cloud.domainedessablons.fr/remote.php/dav/',
        'password': 'VeyshleewdmynJig7ob',
        'username': 'tobixen',
        'incompatibilities': compatibility_issues.nextcloud + ['unique_calendar_ids', 'non_existing_calendar_found', 'sticky_events'], ## the two latter seems to be a sticky calendar thing and/or impossible to create two calendars with the same name
        'enable': True # not working 100% as of 2022-04 ?
    },
    { # 2
        'url': 'https://ecloud.global/remote.php/dav/',
        'username': 'tobixen@e.email',
        'password': '/Test4Caldav.',
        'incompatibilities': compatibility_issues.nextcloud + ['unique_calendar_ids', 'non_existing_calendar_found'],
        'enable': True
    },
    { # 3
        ## <gregn@fastmail.com>
        'url': 'https://caldav.luckydrawcalendar.com/dav.php',
        'username': 'alpha',
        'password': 'passwordpassword',
        'incompatibilities': compatibility_issues.baikal + [ 'dav_not_supported' ],
        'enable': True ## getting only 401 after 2022-05-24, contacted greg at 2022-05-29
    },
    { # 4
        'url': 'https://zimbra.redpill-linpro.com/dav/',
        'username': 'tobias@redpill-linpro.com',
        'password': 'fiskesuppe kyllingsalat',
        'incompatibilities': compatibility_issues.zimbra,
        'enable': True
    },
    { # 5
        ## Does not get a proper response on get-current-principal propfind.
        ## should do more research into it, maybe
        'url': 'http://dav.cladmi.eu/cal.php',
        'principal_url': '/cal.php/principals/tbrox/',
        'username': 'tbrox',
        'password': 'testdigestauthenticationtbrox',
        'incompatibilities': [ 'no-current-user-principal' ], ## which is very strange, because from the web browser it does show this property
        'enable': False ## mkcalendar fails, calendar.events() fails ... but login works.  Should probably check more on this.  Hm.  It's some sabre thing, and it's possible to create new calendar through the web ui.  weird.
    },
    { # 6
        'url': 'http://calendar.tobixen.no:80/caldav.php/',
        'username': 'tobias',
        'password': '/family.',
        'enable': False, ## temporarily(?) down
        'incompatibilities': compatibility_issues.davical + [ 'no_freebusy_rfc4791' ]
    },

    ## It's needed with manual intervention to refresh accounts:
    { # 7
        'url': 'https://demo2.nextcloud.com/remote.php/dav/',
        'username': 'RQ6PbYE6KaNpfmHY',
        'password': 'demo2',
        'incompatibilities': compatibility_issues.nextcloud,
        'enable': False
    },
    { # 8
        'url': 'https://caldav.fastmail.com/dav/',
        'username': 'pythoncaldavtest@fastmail.com',
        'password': 'Q66M664VUQVJQ4ZJ',
        'incompatibilities': compatibility_issues.fastmail,
        'enable': False
    },
    { #9 - ref github issue #191 - swift.joy7282@fastmail.com
        'url': 'https://calendar.jeanes.us/caldav/',
        'username': 'tobixen',
        'password': 'skiing-chewer-tipping-sarcasm-knee',
        'incompatibilities': compatibility_issues.fastmail,
    },
    { #10 - https://calendar.robur.coop/ - ref https://github.com/python-caldav/caldav/issues/213
        'url': 'https://calendar.robur.coop/principals/',
        'username': 'caldavtest',
        'password': 'tycteucbok',
        'incompatibilities': compatibility_issues.robur
    }
]

if False:
    test_xandikos = True
    test_radicale = True
elif False:
    caldav_servers = [{'url': 'http://localhost:8080', 'username': 'user', 'password': 'nopass'}]
elif False:
    for x in caldav_servers:
        x['enable'] = False
    caldav_servers[5]['enable']=True

    test_xandikos = False
    test_radicale = False
else:
    caldav_servers = []
    test_xandikos = True
    test_radicale = True


#caldav_servers = caldav_servers[1:2]
rfc6638_users = []
#for user in ('alpha', 'bravo', 'delta'):
#    rfc6638_users.append({'url': 'https://caldav.luckydrawcalendar.com/dav.php', 'username': user, 'password': 'passwordpassword'})

## SOGo virtual test server
## I did roughly those steps to set up a SOGo test server:
## 1) I download the ZEG - "Zero Effort Groupware" - from https://sourceforge.net/projects/sogo-zeg/
## 2) I installed virtualbox on my laptop
## 3) "virtualbox ~/Downloads/ZEG-5.0.0.ova" (TODO: probably it's possible to launch it "headless"?)
## 4) I clicked on some buttons to get the file "imported" and started
## 5) I went to "tools" -> "preferences" -> "network" and created a NatNetwork
## 6) I think I went to ZEG -> Settings -> Network and chose "Host-only Adapter"
## 7) SOGo was then available at http://192.168.56.101/ from my laptop
## 8) I added the lines below to my conf_private.py
#caldav_servers.append({
#    'url': 'http://192.168.56.101/SOGo/dav/',
#    'username': 'sogo1'.
#    'password': 'sogo'
#})
#for i in (1, 2, 3):
#    sogo = caldav_servers[-1].copy()
#    sogo['username'] = 'sogo%i' % i
#    rfc6638_users.append(sogo)

## MASTER SWITCHES FOR TEST SERVER SETUP
## With those configuration switches, pre-configured test servers in conf.py
## can be turned on or off

## test_public_test_servers - Use the list of common public test
## servers from conf.py.  As of 2020-10 no public test servers exists, so this option
## is currently moot :-(
test_public_test_servers = False

## test_private_test_servers - test using the list configured above in this file.
test_private_test_servers = True

## For usage by ../examples/scheduling_examples.py.  Should typically
## be three different users on the same caldav server.
#rfc6638_users = [ caldav_servers[0], caldav_servers[1], caldav_servers[2] ]



from six import PY3
from six import string_types


def isPython3():
    """Deprecated. Use six.PY3"""
    return PY3


def to_wire(text):
    if text is None:
        return None
    if isinstance(text, string_types) and PY3:
        text = bytes(text, "utf-8")
    elif not PY3:
        text = to_unicode(text).encode("utf-8")
    text = text.replace(b"\n", b"\r\n")
    text = text.replace(b"\r\r\n", b"\r\n")
    return text


def to_local(text):
    if text is None:
        return None
    if not isinstance(text, string_types):
        text = text.decode("utf-8")
    text = text.replace("\r\n", "\n")
    return text


to_str = to_local


def to_normal_str(text):
    """
    A str object is a unicode on python3 and a byte string on python2.
    Make sure we return a normal string, no matter what version of
    python ...
    """
    if text is None:
        return text
    if PY3 and not isinstance(text, str):
        text = text.decode("utf-8")
    elif not PY3 and not isinstance(text, str):
        text = text.encode("utf-8")
    text = text.replace("\r\n", "\n")
    return text


def to_unicode(text):
    if (
        text
        and isinstance(text, string_types)
        and not PY3
        and not isinstance(text, unicode)
    ):
        return unicode(text, "utf-8")
    if PY3 and text and isinstance(text, bytes):
        return text.decode("utf-8")
    return text



#!/usr/bin/python
# -*- encoding: utf-8 -*-
import sys

from setuptools import find_packages
from setuptools import setup

## ATTENTION! when doing releases, the default debugmode in lib/error.py should be set to PRODUCTION.
## (TODO: any nicer ways than doing this manually?  Make a "releases" branch, maybe?)
version = "0.11.0"

if __name__ == "__main__":
    ## For python 2.7 and 3.5 we depend on pytz and tzlocal.  For 3.6 and up, batteries are included.  Same with mock. (But unfortunately the icalendar library only support pytz timezones, so we'll keep pytz around for a bit longer).
    try:
        import datetime
        from datetime import timezone

        datetime.datetime.now().astimezone(timezone.utc)
        extra_packages = []
        ## line below can be removed when https://github.com/collective/icalendar/issues/333 is fixed
        extra_packages = ["pytz", "tzlocal"]
    except:
        extra_packages = ["pytz", "tzlocal"]
    try:
        from unittest.mock import MagicMock

        extra_test_packages = []
    except:
        extra_test_packages = ["mock"]

    test_packages = [
        "pytest",
        "pytest-coverage",
        "icalendar",
        "coverage",
        "tzlocal",
        "pytz",
        "xandikos",
        "radicale",
    ]

    setup(
        name="caldav",
        version=version,
        py_modules=[
            "caldav",
        ],
        description="CalDAV (RFC4791) client library",
        long_description=open("README.md").read(),
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: GNU General " "Public License (GPL)",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Topic :: Office/Business :: Scheduling",
            "Topic :: Software Development :: Libraries " ":: Python Modules",
        ],
        keywords="",
        author="Cyril Robert",
        author_email="cyril@hippie.io",
        url="https://github.com/python-caldav/caldav",
        license="GPL",
        packages=find_packages(exclude=["tests"]),
        include_package_data=True,
        zip_safe=False,
        install_requires=[
            "vobject",
            "lxml",
            "requests",
            "six",
            "icalendar",
            "recurring-ical-events>=1.1.0b",
        ]
        + extra_packages,
        tests_require=test_packages + extra_test_packages,
        extras_require={
            "test": test_packages,
        },
    )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import datetime
import re
import uuid

import icalendar
from caldav.lib.python_utilities import to_normal_str

## Fixups to the icalendar data to work around compatbility issues.

## TODO:

## 1) this should only be done if needed.  Use try-except around the
## fragments where icalendar/vobject is parsing ical data, and do the
## fixups there.

## 2) arguably, this is outside the scope of the caldav library.
## check if this can be done in vobject or icalendar libraries instead
## of here

## TODO: would be nice with proper documentation on what systems are
## generating broken data.  Compatibility issues should also be collected
## in the documentation. somewhere.
def fix(event):
    """This function receives some ical as it's given from the server, checks for
    breakages with the standard, and attempts to fix up known issues:

    1) COMPLETED MUST be a datetime in UTC according to the RFC, but sometimes
    a date is given. (Google Calendar?)

    2) The RFC does not specify any range restrictions on the dates,
    but clearly it doesn't make sense with a CREATED-timestamp that is
    centuries or decades before RFC2445 was published in 1998.
    Apparently some calendar servers generate nonsensical CREATED
    timestamps while other calendar servers can't handle CREATED
    timestamps prior to 1970.  Probably it would make more sense to
    drop the CREATED line completely rather than moving it from the
    end of year 0AD to the beginning of year 1970. (Google Calendar)

    3) iCloud apparently duplicates the DTSTAMP property sometimes -
    keep the first DTSTAMP encountered (arguably the DTSTAMP with earliest value
    should be kept).

    4) ref https://github.com/python-caldav/caldav/issues/37,
    X-APPLE-STRUCTURED-EVENT attribute sometimes comes with trailing
    white space.  I've decided to remove all trailing spaces, since
    they seem to cause a traceback with vobject and those lines are
    simply ignored by icalendar.
    """
    ## TODO: add ^ before COMPLETED and CREATED?
    ## 1) Add a random time if completed is given as date
    fixed = re.sub(
        r"COMPLETED:(\d+)\s", r"COMPLETED:\g<1>T120000Z", to_normal_str(event)
    )

    ## 2) CREATED timestamps prior to epoch does not make sense,
    ## change from year 0001 to epoch.
    fixed = re.sub("CREATED:00001231T000000Z", "CREATED:19700101T000000Z", fixed)
    fixed = re.sub(r"\\+('\")", r"\1", fixed)

    ## 4) trailing whitespace probably never makes sense
    fixed = re.sub(" *$", "", fixed)

    ## 3 fix duplicated DTSTAMP
    ## OPTIMIZATION TODO: use list and join rather than concatination
    ## remove duplication of DTSTAMP
    fixed2 = ""
    for line in fixed.strip().split("\n"):
        if line.startswith("BEGIN:V"):
            cnt = 0
        if line.startswith("DTSTAMP:"):
            if not cnt:
                fixed2 += line + "\n"
            cnt += 1
        else:
            fixed2 += line + "\n"

    return fixed2


## sorry for being english-language-euro-centric ... fits rather perfectly as default language for me :-)
def create_ical(ical_fragment=None, objtype=None, language="en_DK", **props):
    """
    I somehow feel this fits more into the icalendar library than here
    """
    ical_fragment = to_normal_str(ical_fragment)
    if not ical_fragment or not re.search("^BEGIN:V", ical_fragment, re.MULTILINE):
        my_instance = icalendar.Calendar()
        my_instance.add("prodid", "-//python-caldav//caldav//" + language)
        my_instance.add("version", "2.0")
        if objtype is None:
            objtype = "VEVENT"
        component = icalendar.cal.component_factory[objtype]()
        component.add("dtstamp", datetime.datetime.now())
        component.add("uid", uuid.uuid1())
        my_instance.add_component(component)
    else:
        if not ical_fragment.startswith("BEGIN:VCALENDAR"):
            ical_fragment = (
                "BEGIN:VCALENDAR\n"
                + to_normal_str(ical_fragment.strip())
                + "\nEND:VCALENDAR\n"
            )
        my_instance = icalendar.Calendar.from_ical(ical_fragment)
        component = my_instance.subcomponents[0]
        ical_fragment = None
    for prop in props:
        if props[prop] is not None:
            if prop in ("child", "parent"):
                for value in props[prop]:
                    component.add(
                        "related-to", props[prop], parameters={"rel-type": prop.upper()}
                    )
            else:
                component.add(prop, props[prop])
    ret = to_normal_str(my_instance.to_ical())
    if ical_fragment and ical_fragment.strip():
        ret = re.sub(
            "^END:V", ical_fragment.strip() + "\nEND:V", ret, flags=re.MULTILINE
        )
    return ret



#!/usr/bin/env python
import re
import uuid
from datetime import datetime
from datetime import timedelta
from unittest import TestCase

import icalendar
import pytz
import vobject
from caldav.lib.python_utilities import to_normal_str
from caldav.lib.python_utilities import to_wire
from caldav.lib.vcal import create_ical
from caldav.lib.vcal import fix

# from datetime import timezone
# utc = timezone.utc
utc = pytz.utc

# example from http://www.rfc-editor.org/rfc/rfc5545.txt
ev = """BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//Example Corp.//CalDAV Client//EN
BEGIN:VEVENT
UID:19970901T130000Z-123403@example.com
DTSTAMP:19970901T130000Z
DTSTART;VALUE=DATE:19971102
SUMMARY:Our Blissful Anniversary
TRANSP:TRANSPARENT
CLASS:CONFIDENTIAL
CATEGORIES:ANNIVERSARY,PERSONAL,SPECIAL OCCASION
RRULE:FREQ=YEARLY
END:VEVENT
END:VCALENDAR"""


class TestVcal(TestCase):
    def assertSameICal(self, ical1, ical2, ignore_uid=False):
        """helper method"""

        def normalize(s, ignore_uid):
            s = to_wire(s).replace(b"\r\n", b"\n").strip().split(b"\n")
            s.sort()
            if ignore_uid:
                s = [x for x in s if not x.startswith(b"UID:")]
            return b"\n".join(s)

        self.assertEqual(normalize(ical1, ignore_uid), normalize(ical2, ignore_uid))
        return ical2

    def verifyICal(self, ical):
        """
        Does a best effort on verifying that the ical is correct, by
        pushing it through the vobject and icalendar library
        """
        vobj = vobject.readOne(to_normal_str(ical))
        icalobj = icalendar.Calendar.from_ical(ical)
        self.assertSameICal(icalobj.to_ical(), ical)
        self.assertSameICal(vobj.serialize(), ical)
        return icalobj.to_ical()

    ## TODO: create a test_fix, should be fairly simple - for each
    ## "fix" that's done in the code, make up some broken ical data
    ## that demonstrates the brokenness we're dealing with (preferably
    ## real-world examples). Then ...
    # for bical in broken_ical:
    #    verifyICal(vcal.fix(bical))

    def test_create_ical(self):
        def create_and_validate(**args):
            return self.verifyICal(create_ical(**args))

        ## First, a fully valid ical_fragment should go through as is
        self.assertSameICal(create_and_validate(ical_fragment=ev), ev)

        ## One may add stuff to a fully valid ical_fragment
        self.assertSameICal(
            create_and_validate(ical_fragment=ev, priority=3), ev + "\nPRIORITY:3\n"
        )

        ## binary string or unicode string ... shouldn't matter
        self.assertSameICal(
            create_and_validate(ical_fragment=ev.encode("utf-8"), priority=3),
            ev + "\nPRIORITY:3\n",
        )

        ## The returned ical_fragment should always contain BEGIN:VCALENDAR and END:VCALENDAR
        ical_fragment = ev.replace("BEGIN:VCALENDAR", "").replace("END:VCALENDAR", "")
        self.assertSameICal(create_and_validate(ical_fragment=ical_fragment), ev)

        ## Create something with a dtstart and verify that we get it back in the ical
        some_ical0 = create_and_validate(
            summary="gobledok",
            dtstart=datetime(2032, 10, 10, 10, 10, 10, tzinfo=utc),
            duration=timedelta(hours=5),
        )
        some_ical1 = create_and_validate(
            summary=b"gobledok",
            dtstart=datetime(2032, 10, 10, 10, 10, 10, tzinfo=utc),
            duration=timedelta(hours=5),
        )
        assert re.search(b"DTSTART(;VALUE=DATE-TIME)?:20321010T101010Z", some_ical0)
        self.assertSameICal(some_ical0, some_ical1, ignore_uid=True)

        ## Verify that ical_fragment works as intended
        some_ical = create_and_validate(
            summary="gobledok",
            ical_fragment="PRIORITY:3",
            dtstart=datetime(2032, 10, 10, 10, 10, 10, tzinfo=utc),
            duration=timedelta(hours=5),
        )
        assert re.search(b"DTSTART(;VALUE=DATE-TIME)?:20321010T101010Z", some_ical)

        some_ical = create_and_validate(
            summary="gobledok",
            ical_fragment=b"PRIORITY:3",
            dtstart=datetime(2032, 10, 10, 10, 10, 10, tzinfo=utc),
            duration=timedelta(hours=5),
        )
        assert re.search(b"DTSTART(;VALUE=DATE-TIME)?:20321010T101010Z", some_ical)

        some_ical = create_and_validate(
            summary=b"gobledok",
            ical_fragment="",
            dtstart=datetime(2032, 10, 10, 10, 10, 10, tzinfo=utc),
            duration=timedelta(hours=5),
        )
        assert re.search(b"DTSTART(;VALUE=DATE-TIME)?:20321010T101010Z", some_ical)
